(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010 Steffen Solyga <solyga@absinth.net>                               *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Gen IMPORT TValue, TIndex, TAddress, TType, TMode, TItemDesc;
FROM Gen IMPORT Hlt, Wrt, Op1, Op2, Asg;
FROM Gen IMPORT Label, Init, Jmp, Fixup, JmpFalse, FixFalse, JmpFalseBwd;
FROM Gen IMPORT vFalse, vTrue;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TClass = ( cConst, cVar, cType );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: TIndex;
    | cType :
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;

PROCEDURE MkItem( VAR x: TItemDesc; o: TObjDesc );
BEGIN
  x.t := o.typ;
  CASE o.cls OF
  | cConst: x.m := mConst; x.v := o.val;
  | cVar  : x.m := mVar;   x.i := o.idx;
  ELSE
    Err( 'Neither constant nor variable' );
  END;
END MkItem;

PROCEDURE MkConstItem( VAR x: TItemDesc; t: TType; v: TValue );
BEGIN
  x.t := t; x.m := mConst; x.v := v;
END MkConstItem;

PROCEDURE ChkBool( VAR x: TItemDesc );
BEGIN
  IF x.t # tBool THEN Err( 'Operand not boolean' ) END;
END ChkBool;

PROCEDURE ChkInt( VAR x: TItemDesc );
BEGIN
  IF x.t # tInt THEN Err( 'Operand not integer' ) END;
END ChkInt;

PROCEDURE ChkCompat( VAR x, y: TItemDesc );
BEGIN
  IF x.t # y.t THEN Err( 'Incompatible types' ) END;
END ChkCompat;


PROCEDURE Expression( VAR x: TItemDesc );

  PROCEDURE SimpleExpr( VAR x: TItemDesc );

    PROCEDURE Factor( VAR x: TItemDesc );
    VAR i: CARDINAL;
    BEGIN
      CASE sym OF
      | sNum : IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
               MkConstItem( x, tInt, num ); Read();
      | sId  : Find( i );
               IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
               MkItem( x, objs[i] ); Read();
      | sLPar: Read(); Expression( x );
               IF sym # sRPar THEN Err( 'Expected operator or ")"' ) END;
               Read();
      | sNot : Read(); Factor( x ); ChkBool( x ); Op1( sNot, x );
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR x: TItemDesc );
    VAR s: TSymbol; y: TItemDesc;
    BEGIN
      Factor( x );
      WHILE ( sMul <= sym ) & ( sym <= sAnd ) DO
        IF sym = sAnd THEN ChkBool( x ); Op1( sAnd, x ) ELSE ChkInt( x ) END;
        s := sym; Read();
        Factor( y ); ChkCompat( x, y ); Op2( s, x, y );
      END;
    END Term;

  VAR s: TSymbol; y: TItemDesc;
  BEGIN
    CASE sym OF
    | sPlus : Read(); Term( x ); ChkInt( x );
    | sMinus: Read(); Term( x ); ChkInt( x ); Op1( sMinus, x );
    ELSE
      Term( x );
    END;
    WHILE ( sPlus <= sym ) & ( sym <= sOr ) DO
      IF sym = sOr THEN ChkBool( x ); Op1( sOr, x ) ELSE ChkInt( x ) END;
      s := sym; Read();
      Term( y ); ChkCompat( x, y ); Op2( s, x, y );
    END;
  END SimpleExpr;

VAR s: TSymbol; y: TItemDesc;
BEGIN
  SimpleExpr( x );
  IF ( sEql <= sym ) & ( sym <= sGeq ) THEN
    Op1( sym, x );
    s := sym; Read();
    SimpleExpr( y ); ChkCompat( x, y ); Op2( s, x, y );
    x.t := tBool;
  END;
END Expression;

PROCEDURE ConstDeclaration();
VAR i: CARDINAL; x: TItemDesc;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cConst;
  KeepId();
  Read();
  IF sym # sEql THEN Err( 'Expected "="' ) END;
  Read(); Expression( x );
  IF x.m # mConst THEN Err( 'Expression not constant' ) END;
  objs[ i ].typ := x.t;
  objs[ i ].val := x.v;
  INC( cnt );
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END ConstDeclaration;

PROCEDURE VarDeclaration();
VAR i,j: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cVar;
  KeepId();
  Read();
  IF sym # sCol THEN Err( 'Expected ":"' ) END;
  Read();
  IF sym # sId THEN Err( 'Expected identifier' ) END;
  Find( j );
  IF j >= cnt THEN Err( 'Identifier undeclared' ) END;
  IF objs[ j ].cls # cType THEN Err( 'Not a type' ) END;
  IF idx > MAX( TIndex ) THEN Err( 'Too many variables' ) END;
  objs[ i ].typ := objs[ j ].typ;
  objs[ i ].idx := idx;
  INC( cnt ); INC( idx );
  Read();
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END VarDeclaration;

PROCEDURE StatementSequence();

  PROCEDURE Statement();
  VAR i: CARDINAL; x,y: TItemDesc; l: TAddress;
  BEGIN
    CASE sym OF
    | sId: (* assignment *)
        Find( i );
        IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
        IF objs[ i ].cls # cVar THEN Err( 'Not a variable' ) END;
        MkItem( x, objs[i] ); Read();
        IF sym # sBec THEN Err( 'Expected ":="' ) END;
        Read(); Expression( y ); ChkCompat( x, y ); Asg( x, y );
    | sWrt: (* expression output *)
        Read(); Expression( x ); Wrt( x );
    | sRepeat:
        Read(); Label( l ); StatementSequence();
        IF sym # sUntil THEN Err( 'Expected "UNTIL"' ) END;
        Read(); Expression( x ); ChkBool( x ); JmpFalseBwd( l, x );
    | sWhile:
        Read(); Label( l ); Expression( x ); ChkBool( x );
        IF sym # sDo THEN Err( 'Expected "DO"' ) END;
        Read(); JmpFalse( x );
        StatementSequence(); Jmp( l ); FixFalse( x );
        IF sym # sEnd THEN Err( 'Expected "END"' ) END;
        Read();
    | sIf:
        Read(); Expression( x ); ChkBool( x ); JmpFalse( x );
        IF sym # sThen THEN Err( 'Expected "THEN"' ) END;
        Read(); StatementSequence();
        Init( l );
        WHILE sym = sElsif DO
          Read(); Jmp( l ); FixFalse( x );
          Expression( x ); ChkBool( x ); JmpFalse( x );
          IF sym # sThen THEN Err( 'Expected "THEN"' ) END;
          Read(); StatementSequence();
        END;
        IF sym = sElse THEN
          Read(); Jmp( l ); FixFalse( x ); StatementSequence();
        ELSE
          FixFalse( x );
        END;
        Fixup( l );
        IF sym # sEnd THEN Err( 'Expected "END"' ) END;
        Read();
    ELSE
      (* empty *)
    END;
  END Statement;

BEGIN
  Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
END StatementSequence;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read();
  IF sym = sConst THEN
    Read();
    WHILE sym = sId DO ConstDeclaration() END;
  END;
  IF sym = sVar THEN
    Read();
    WHILE sym = sId DO VarDeclaration() END;
  END;
  IF sym # sBeg THEN Err( 'Expected "BEGIN"' ) END;
  Read(); StatementSequence();
  IF sym # sEnd THEN Err( '"BEGIN" requires "END"' ) END;
  Read();
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
(*  Read(); IF sym # sEOF THEN Err( 'Expected EOF' ) END; *)
  Hlt();
END Run;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vFalse;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vTrue;
  KeepId(); INC( cnt );
  Put( 'BOOLEAN' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  Put( 'INTEGER' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tInt;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
END Prsr.
